from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

USER_SETTINGS = getattr(settings, "REST_FRAMEWORK_MICROSERVICE", None)

DEFAULTS = {
    "REFRESH_COOKIE_NAME": "refresh_cookie",
    "REFRESH_COOKIE_PATH": "/auth",
    "IDP": {
        "PROVIDER": "aws",
        "REGION": "us-west-2",
        "USER_POOL": "us-west-2_abcdefg",
        "CLIENT_ID": "abcdefg",
    },
    "CUSTOM_TOKEN_USER_ATTRIBUTES": [],
    "CUSTOM_TOKEN_CALLABLE_ATTRIBUTES": [],
    "COOKIE_SALT": "extra",
    "USER_SERIALIZER_CLASS": None,
    "USER_MODEL_UUID_FIELD": None,
}

IMPORT_STRINGS = [
    'USER_SERIALIZER_CLASS'
]


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class RestFrameworkMicroserviceSettings:
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        self._user_settings = user_settings or {}
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "REST_FRAMEWORK_MICROSERVICE", {})
        return self._user_settings

    def __getattr__(self, attr):

        # check the setting is accepted attribute
        if attr not in self.defaults:
            raise AttributeError(f"Invalid REST_FRAMEWORK_MICROSERVICE setting: {attr}")

        # get from user settings or default value
        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


rest_microservice_settings = RestFrameworkMicroserviceSettings(USER_SETTINGS, DEFAULTS)


def reload_rest_framework_microservice_settings(*args, **kwargs):
    print("Reloading rest_framework_microservice settings")
    setting = kwargs["setting"]
    if setting == "REST_FRAMEWORK_MICROSERVICE":
        rest_microservice_settings.reload()


setting_changed.connect(reload_rest_framework_microservice_settings)



from datetime import datetime
from django.conf import settings
from .settings import rest_microservice_settings
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework_simplejwt.views import TokenObtainPairView, TokenViewBase

from .serializers import LogInTokenObtainPairSerializer, CustomTokenRefreshSerializer, \
    SocialLogInTokenExchangeSerializer


class RefreshTokenUsingCookieMixin:
    """
    Provides function to handle refresh token that comes from cookie
    """

    def set_cookie_header_in_response(self, response, refresh_token, refresh_expiry):
        expires = datetime.fromtimestamp(refresh_expiry)

        response.set_signed_cookie(key=rest_microservice_settings.REFRESH_COOKIE_NAME,
                                   value=refresh_token,
                                   salt=rest_microservice_settings.COOKIE_SALT,
                                   expires=expires,
                                   httponly=True, samesite='strict', secure=not settings.DEBUG,
                                   path=rest_microservice_settings.REFRESH_COOKIE_PATH)
        return response

    @staticmethod
    def get_token_from_cookie(request):
        try:
            token = request.get_signed_cookie(rest_microservice_settings.REFRESH_COOKIE_NAME,
                                              salt=rest_microservice_settings.COOKIE_SALT)
        except KeyError as e:
            raise InvalidToken()

        return token

    @staticmethod
    def get_delete_cookie_response(status_code=status.HTTP_401_UNAUTHORIZED):
        response = Response(status=status_code)
        response.delete_cookie(key=rest_microservice_settings.REFRESH_COOKIE_NAME,
                               path=rest_microservice_settings.REFRESH_COOKIE_PATH)
        return response


class TokenLogIn(TokenObtainPairView, RefreshTokenUsingCookieMixin):
    """
    Log in using username and password, returns access token in body, refresh token in httpOnly cookie.
    """

    serializer_class = LogInTokenObtainPairSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)

        try:
            serializer.is_valid(raise_exception=True)
        except TokenError as e:
            raise InvalidToken(e.args[0])

        refresh_token = serializer.validated_data['refresh']
        refresh_expiry = serializer.validated_data['refresh_expiry']
        response = serializer.make_auth_response()

        return self.set_cookie_header_in_response(response, refresh_token, refresh_expiry)


class RefreshTokenUsingCookie(TokenViewBase, RefreshTokenUsingCookieMixin):
    """
    Provides an access token when called with a refresh token in header cookie.
    """
    serializer_class = CustomTokenRefreshSerializer

    def post(self, request, *args, **kwargs):
        jwt = self.get_token_from_cookie(request)
        serializer = self.get_serializer(data={'refresh': jwt})

        if not serializer.is_valid():
            return self.get_delete_cookie_response()

        response = serializer.make_auth_response()
        return response


class BlacklistRefreshToken(TokenViewBase, RefreshTokenUsingCookieMixin):
    """
    Blacklist's refresh token received in header cookie. Used for logging off.
    """

    def post(self, request, *args, **kwargs):
        jwt = self.get_token_from_cookie(request)
        token = RefreshToken(jwt)
        token.blacklist()
        return self.get_delete_cookie_response(status_code=status.HTTP_200_OK)


class SocialLogInExchangeTokens(APIView, RefreshTokenUsingCookieMixin):
    """
    Used when front-end authenticates directly with auth provider using OAuth2 Code grant with PKCE.
    This end point allows frontend to exchange auth provider JWT for backend-signed tokens.
    """
    permission_classes = ()
    authentication_classes = ()

    def post(self, request, *args, **kwargs):
        serializer = SocialLogInTokenExchangeSerializer(data=request.data)

        serializer.is_valid(raise_exception=True)
        refresh_token = serializer.validated_data['refresh']
        refresh_expiry = serializer.validated_data['refresh_expiry']
        response = serializer.make_auth_response()

        return self.set_cookie_header_in_response(response, refresh_token, refresh_expiry)



from rest_framework.exceptions import APIException
from rest_framework import status
from django.utils.translation import gettext_lazy as _


class ModelAttributeError(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = _('Model instance attribute error.')
    default_code = 'authentication_failed'


class TokenExpired(APIException):
    status_code = status.HTTP_410_GONE
    default_detail = _('Token expired.')
    default_code = 'authentication_failed'


class InvalidToken(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _('Invalid token.')
    default_code = 'authentication_failed'


class SubscriptionExpired(APIException):
    status_code = status.HTTP_402_PAYMENT_REQUIRED
    default_detail = _('Subscription expired.')
    default_code = 'error'


class UserAgreementExpired(APIException):
    status_code = status.HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS
    default_detail = _('User agreement expired.')
    default_code = 'error'



from operator import attrgetter
from django.conf import settings
from .settings import rest_microservice_settings
from rest_framework import serializers
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework_simplejwt.serializers import TokenObtainSerializer, TokenRefreshSerializer
from rest_framework_simplejwt.tokens import RefreshToken
from rest_framework_simplejwt.settings import api_settings
from django.utils.module_loading import import_string
from .exceptions import InvalidToken, TokenExpired
from django.contrib.auth.models import update_last_login
from django.contrib.auth import get_user_model
from jwt import ExpiredSignatureError
from django.core.exceptions import ObjectDoesNotExist
from rest_framework.response import Response
from .social_auth.aws_cognito import decode_token
from .models import Idp

def add_custom_token_claims(token, user):
    """Append custom token claims specified using settings."""
    for attr in rest_microservice_settings.CUSTOM_TOKEN_USER_ATTRIBUTES:
        token[attr] = getattr(user, attr, None)

    for attr in rest_microservice_settings.CUSTOM_TOKEN_CALLABLE_ATTRIBUTES:
        attr_name = attr["attr_name"]
        attr_getter = import_string(attr["attr_getter"])
        token[attr_name] = attr_getter(user)

    return token


class SerializerResponseMixin:
    def make_auth_response(self):
        access_token = self.validated_data['access']
        access_expiry = self.validated_data['access_expiry']
        response_data = {'access_token': access_token, 'expires': access_expiry}

        user_serializer = rest_microservice_settings.USER_SERIALIZER_CLASS
        if user_serializer:
            user = self.validated_data['user']
            response_data['user'] = user_serializer(user).data

        return Response(response_data, status=status.HTTP_200_OK)


class LogInTokenObtainPairSerializer(TokenObtainSerializer, SerializerResponseMixin):
    """
    Validates user, returns dictionary containing refresh token, access token,
    expiry time of access token, user dictionary.
    This is similar to the rest_framework_simplejwt.serializers.TokenObtainPairSerializer.
    """

    @classmethod
    def get_token(cls, user):
        # TODO: simple-jwt future release
        # token = cls.token_class.for_user(user)
        token = RefreshToken.for_user(user)
        add_custom_token_claims(token, user)

        return token

    def validate(self, attrs):
        data = super().validate(attrs)

        refresh = self.get_token(self.user)

        data['refresh'] = str(refresh)
        data['access'] = str(refresh.access_token)

        # putting a few things from JWT directly into the response to make it easier for web frontend to use
        data['refresh_expiry'] = refresh.payload['exp']
        data['access_expiry'] = refresh.access_token['exp']
        data['user'] = self.user

        update_last_login(None, self.user)

        return data


class CustomTokenRefreshSerializer(TokenRefreshSerializer, SerializerResponseMixin):
    """
    Validates refresh token string, returns dictionary containing access token,
    expiry time of access token, user dictionary.
    """

    def validate(self, attrs):
        try:
            refresh = RefreshToken(attrs['refresh'])
            user = self.get_user(refresh)
            # copying custom attributes from user instance to TokenUser in case any attributes have been changed
            add_custom_token_claims(refresh, user)

        # only intercept errors that we do not want to see in Django error reporting here
        except ExpiredSignatureError:
            raise TokenExpired()
        except ObjectDoesNotExist:
            raise InvalidToken()

        data = {'access': str(refresh.access_token), 'access_expiry': refresh.access_token['exp'], 'user': user}

        if api_settings.ROTATE_REFRESH_TOKENS:
            if api_settings.BLACKLIST_AFTER_ROTATION:
                try:
                    # Attempt to blacklist the given refresh token
                    refresh.blacklist()
                except AttributeError:
                    # If blacklist app not installed, `blacklist` method will not be present
                    pass

            refresh.set_jti()
            refresh.set_exp()

            data['refresh_token'] = str(refresh)

        return data

    @staticmethod
    def get_user(refresh_token):
        user_model = get_user_model()
        return user_model.objects.get(id=refresh_token.get("user_id"))


class SocialLogInTokenExchangeSerializer(serializers.Serializer, SerializerResponseMixin):
    """
    Validates AWS Cognito id token, returns dictionary containing internally-signed
    refresh token, access token, expiry time of access token, and user dictionary.
    """
    id_token = serializers.CharField()
    access_token = serializers.CharField()
    refresh_token = serializers.CharField()

    @classmethod
    def get_token(cls, user):
        return RefreshToken.for_user(user)

    def validate(self, attrs):
        try:
            id_token = decode_token(attrs['id_token'], rest_microservice_settings.IDP['CLIENT_ID'])

        # only intercept errors that we do not want to see in Django error report here
        except ExpiredSignatureError:
            raise TokenExpired()

        if id_token.get('email_verified') is False:
            raise APIException("Email not verified.")

        user_defaults = {
            "username": id_token["cognito:username"],
            "first_name": id_token.get("given_name"),
            "last_name": id_token.get("family_name"),
        }

        if rest_microservice_settings.USER_MODEL_UUID_FIELD is not None:
            user_defaults[rest_microservice_settings.USER_MODEL_UUID_FIELD] = id_token['sub']

        user, created = get_user_model().objects.get_or_create(email=id_token["email"], defaults=user_defaults)

        if rest_microservice_settings.USER_MODEL_UUID_FIELD is None:
            Idp.objects.update_or_create(user=user, defaults={"uuid": id_token['sub']})

        refresh = self.get_token(user)
        data = {
            'refresh': str(refresh),
            'access': str(refresh.access_token),
            'refresh_expiry': refresh.payload['exp'],
            'access_expiry': refresh.access_token['exp'],
            'user': user
        }

        update_last_login(None, user)

        return data



# Generated by Django 4.0.3 on 2022-04-04 00:00

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Idp',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField()),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='idp', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddIndex(
            model_name='idp',
            index=models.Index(fields=['uuid'], name='rest_framew_uuid_a01bb5_idx'),
        ),
    ]



from rest_framework_simplejwt.models import TokenUser
from django.db import models
from django.contrib.auth import get_user_model


class CustomTokenUser(TokenUser):
    """
    Extend TokenUser and adds custom attributes to be pulled from TokenUser.
    This class should be specified in Django settings SIMPLE_JWT.TOKEN_USER_CLASS
    """

    # TODO: No longer needed in newer version of djangorestframework-simplejwt
    # See: https://github.com/jazzband/djangorestframework-simplejwt/pull/528
    def __getattr__(self, attr_name):
        return self.token.get(attr_name, None)


class Idp(models.Model):
    user = models.OneToOneField(get_user_model(), on_delete=models.CASCADE, related_name="idp")
    uuid = models.UUIDField()

    class Meta:
        indexes = [models.Index(fields=['uuid'])]